from collections.abc import Mapping
from typing import Optional

from flask import Request
from cloudevents.http import CloudEvent


class Context(Mapping):
    """
    Class holding invocation context.

    ...

    Attributes
    ----------
    request:
        Flask HTTP request
    cloud_event:
        CloudEvent if any
    """
    request: Request
    cloud_event: Optional[CloudEvent]
    __slots__ = ['request', 'cloud_event']

    def __init__(self, req: Request):
        self.request = req
        self.cloud_event = None

    # for backward compatibility reason this class can behave like a Mapping

    def __getitem__(self, item):
        if item == 'request' and self.request is not None:
            return self.request
        if item == 'cloud_event' and self.cloud_event is not None:
            return self.cloud_event
        raise KeyError(item)

    def __iter__(self):
        if self.request is not None:
            yield 'request'
        if self.cloud_event is not None:
            yield 'cloud_event'

    def __len__(self):
        return int(self.request is not None) +\
               int(self.cloud_event is not None)



import os
import sys
import traceback

from flask import Flask, request
from cloudevents.http import CloudEvent, from_http, to_binary
from .invocation import Context


def load(path):
    """
    Load a python file with a main() function and return the module
    """
    func_dir = os.path.realpath(path)
    sys.path.append(func_dir)
    import func
    return func


def create(func):
    """
    Create a Flask app with kube health endpoints, exposing 'func' at /
    """
    app = Flask(__name__)

    @app.route("/", methods=["POST"])
    def handle_post():
        context = Context(request)
        try:
            context.cloud_event = from_http(request.headers,
                                            request.get_data())
        except Exception:
            app.logger.warning('No CloudEvent available')
        return invoke(func, context)

    @app.route("/", methods=["GET"])
    def handle_get():
        context = Context(request)
        return invoke(func, context)

    @app.route("/health/liveness")
    def liveness():
        return "OK"

    @app.route("/health/readiness")
    def readiness():
        return "OK"

    return app


def invoke(func, context):
    try:
        resp = func.main(context)
        if isinstance(resp, CloudEvent):
            headers, body = to_binary(resp)
            return body, headers
        else:
            return resp
    except Exception as err:
        traceback.print_exc()
        print("caught", err)
        return f"Function raised {err}", 500



# __main__.py

import sys
import signal

from waitress import serve
from . import server


def main():
    """
    Loads a python function, specified by the first command line
    argument, and serves it using the parliament server.
    """
    if (len(sys.argv) != 2):
        print("Usage: python -m", __file__, "<path/to/func.py>")
        exit(1)
    signal.signal(signal.SIGTERM, receive_signal)
    signal.signal(signal.SIGINT, receive_signal)
    app = server.create(server.load(sys.argv[1]))
    serve(app, host='0.0.0.0', port=8080)


def receive_signal(signal_number, frame):
    sys.exit(128 + signal_number)


if __name__ == "__main__":
    main()



import functools
from cloudevents.http import CloudEvent


def event(_func=None, *,
          event_type="parliament.response",
          event_source="/parliament/function"):
    """
    Provides an @event decorator that accomplishes a few of things.

    1. It converts a function's response value into the data for a CloudEvent
       sending the event back as the response to the caller.

    2. It allows a function developer to specify the CloudEvent
       source and type attributes:

           @event(event_source="/my/source", event_type="my.type")

    3. Defaults the event source and type attributes to "/parliament/function"
       and "parliament.response", respectively.
    """
    def event_decorator(func):
        @functools.wraps(func)
        def event_wrapper(*args, **kwargs):
            data = func(*args, **kwargs)
            if isinstance(data, CloudEvent):
                return data
            attributes = {
              "type": event_type,
              "source": event_source
            }
            return CloudEvent(attributes, data)
        return event_wrapper

    if _func is None:
        return event_decorator
    else:
        return event_decorator(_func)



from setuptools import setup
import pathlib
import os


HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

package_root = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(os.path.join(package_root, "parliament", "version.py")) as fp:
    exec(fp.read(), version)
version = version["__version__"]

setup(
    name='parliament-functions',
    packages=['parliament'],
    version=version,
    license='MIT',
    description='A framework for invoking functions over HTTP',
    long_description=README,
    long_description_content_type='text/markdown',
    author='Lance Ball',
    author_email='lball@redhat.com',
    url='https://github.com/boson-project/parliament',
    keywords=['faas', 'functions', 'openshift'],

    install_requires=['cloudevents', 'flask', 'waitress'],

    classifiers=[],
    entry_points={
      "console_scripts": [
        "parliament=parliament.__main__:main",
      ]
    },
)
